/*
 * Copyright (C) 2009  Thomas "Heinervdm" Zimmermann <zimmermannn@vdm-design.de>
 *
 * This file is part of fmWettbewerb
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.heinervdm.fmwettbewerb;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Return stats:
 *      1: opponent.dat or player.dat not found
 *      2: can not read opponent.dat or player.dat
 *      3: can not write result.dat
 *      4: gamefield.dat not found
 *      5: can not read gamefield.dat
 *      6: no move found
 *
 * @author thomas
 */
public class Main {

    /* Oponent data */
    Player op;
    /* Player data */
    Player me;

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        new Main();
    }

    /**
     * Main function
     */
    public Main() {

        try {
            op = new Player("opponent.dat");
            me = new Player("player.dat");

            Stone[][] playground = initPlayground();

            ArrayList<Move> moves = findMoves(playground);
            Move m = filterMoves(playground, moves);

            writeMove(m);

        } catch (FileNotFoundException ex) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        } catch (IOException ex) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(2);
        }
    }

    /**
     * Finds the best Move
     * @param playground The playground
     * @param moves All possible Moves
     * @return The best move
     */
    private Move filterMoves(Stone[][] playground, ArrayList<Move> moves) {
        if (moves.size() < 1) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, "No move found!");
            System.exit(6);
            return null;
        }

        Move bestMove = null;
        double bestRate = -30.0;
        double bestOpRate = 100.0;

        for (int i = 0; i < moves.size(); i++) {
            Move tmpMove = moves.get(i);
            double tmpRate = rateMove(tmpMove);

            Stone[][] lilaPlayground;

            // if we will get enought lila stones then add the rate for the next move
            if ((tmpMove.getLila() + me.getLila()) >= 15) {
                lilaPlayground = removeRowsFromPlayground(playground, tmpMove.getRows());
                ArrayList<Move> lilaMoves = findMoves(lilaPlayground);
                Move lilaMove = filterOpMoves(lilaMoves);
                tmpRate = rateMove(lilaMove) + tmpRate;
            } else {
                lilaPlayground = playground;
            }

//            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, "Player rate: "+tmpRate);
//            double opRate = 0.0;
            if (tmpRate > bestRate) {
                bestRate = tmpRate;
                bestMove = tmpMove;

                // calculate the the rate for oponents move
                Stone[][] afterPlayground = removeRowsFromPlayground(lilaPlayground, tmpMove.getRows());
                ArrayList<Move> opMoves = findMoves(afterPlayground);
                Move opMove = filterOpMoves(opMoves);
                bestOpRate = rateMove(opMove);

            } else if (tmpRate == bestRate) {
                // if the openents next move is weaker then this will be our best move
                Stone[][] afterPlayground = removeRowsFromPlayground(lilaPlayground, tmpMove.getRows());
                ArrayList<Move> opMoves = findMoves(afterPlayground);
                Move opMove = filterOpMoves(opMoves);
                double opRate = 0.0;
                if (opMove != null) {
                     opRate = rateMove(opMove);
                }

                if (opRate < bestOpRate) {
                    bestRate = tmpRate;
                    bestMove = tmpMove;
                    bestOpRate = opRate;
                }
//                Logger.getLogger(Main.class.getName()).log(Level.SEVERE, "Player rate: " + tmpRate + ", Oponent rate: " + opRate);
            }
        }

        if (bestMove == null) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, "No best move found!");
            System.exit(6);
            return null;
        }

        return bestMove;
    }

    /**
     * Finds the best Move without looking at the playground
     * @param moves All possible Moves
     * @return The best move
     */
    private Move filterOpMoves(ArrayList<Move> moves) {
        if (moves.size() < 1) {
            return null;
        }

        Move bestMove = null;
        double bestRate = 0.0;

        for (int i = 0; i < moves.size(); i++) {
            Move tmpMove = moves.get(i);
            double tmpRate = rateMove(tmpMove);
            if (tmpRate > bestRate) {
                bestRate = tmpRate;
                bestMove = tmpMove;
            }
        }

        return bestMove;
    }

    /**
     * Rates the given Move like follows:
     *      Bombe  = count
     *      Red    = count/15*10
     *      Yellow = count/15*6
     *      Green  = count/15*3
     *      Lila   = count/15
     *      Blue   = count/(15-Shield) || 0.001 if Shield == 15
     * 
     * Farbe 	Bedeutung
     * grün 	sammelt man mindestens 15 Steine, erleidet der Gegner 3 Schaden
     * gelb 	sammelt man mindestens 15 Steine, erleidet der Gegner 6 Schaden
     * rot      sammelt man mindestens 15 Steine, erleidet der Gegner 10 Schaden
     * lila 	sammelt man mindestens 15 Steine, ist man ein zweites Mal an der Reihe
     * blau 	lädt den Schild eines Spielers wieder auf
     *
     * @param playground The playground
     * @param m the Move
     * @return the rating
     */
    private double rateMove(Move m) {
        if (m == null || m.getRows().size() == 0) {
            return -100.0;
        }
        double rate = 0.0;

        ArrayList<Row> rows = m.getRows();

        for (int i = 0; i < rows.size(); i++) {
            Row r = rows.get(i);
            if (Stone.isBomb(r.getColor())) {
                rate += r.getCount();
            } else if (r.getColor().getColor() == Stone.red) {
                rate += ((double) (r.getCount() * 10)) / 15.0;
            } else if (r.getColor().getColor() == Stone.yellow) {
                rate += ((double) (r.getCount() * 6)) / 15.0;
            } else if (r.getColor().getColor() == Stone.green) {
                rate += ((double) (r.getCount() * 3)) / 15.0;
            } else if (r.getColor().getColor() == Stone.lila) {
                rate += ((double) (r.getCount())) / 15.0;
            } else if (r.getColor().getColor() == Stone.blue) {
                if (me.getShield() < 15) {
                    rate += ((double) (r.getCount())) / ((double) (15.0 - me.getShield()));
                } else {
                    rate += 0.01;
                }
            }
        }
        return rate;
    }

    /**
     * Methode for debugging, prints the playground
     * @param tmpPlayground the playground
     */
    private void printPlayground(Stone[][] tmpPlayground) {
        String pImage = "";
        for (int ik = 9; ik >= 0; ik--) {
            for (int jk = 0; jk < 10; jk++) {
                pImage += (tmpPlayground[ik][jk]).getColor();
            }
            pImage += "\r\n";
        }
        Logger.getLogger(Main.class.getName()).log(Level.INFO, "Playground:\r\n" + pImage);
    }

    /**
     * Writes the given Move to result.dat
     * @param m
     */
    private void writeMove(Move m) {
        try {
            FileWriter out = new FileWriter("result.dat");
            out.write(m.getSrcY() + " " + m.getSrcX() + " " + m.getDstY() + " " + m.getDstX());
            out.close();
        } catch (IOException ex) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(3);
        }
    }

    /**
     * Reads the playground into <code>playground</code>
     */
    private Stone[][] initPlayground() {
        Stone[][] playground = new Stone[10][10];
        try {

            BufferedReader in = new BufferedReader(new FileReader("gamefield.dat"));
            for (int i = 9; i >= 0; i--) {
                String line = in.readLine().replaceAll("\\s", "").trim();
                for (int j = 0; j < 10; j++) {
                    playground[i][j] = new Stone(line.substring(j, j + 1).charAt(0));
                }
            }
        } catch (FileNotFoundException ex) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(4);
        } catch (IOException ex) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(5);
        }
        return playground;
    }

    /**
     * finds all posible moves
     * @param playground The playground
     * @return All possible moves
     */
    private ArrayList<Move> findMoves(Stone[][] playground) {
        ArrayList<Move> moves = new ArrayList<Move>();
        for (int i = 0; i < 10; i++) {
            for (int j = 0; j < 10; j++) {
                if (i < 9) {
                    moves.add(swapStones(playground, i, j, i + 1, j));
                }
                if (j < 9) {
                    moves.add(swapStones(playground, i, j, i, j + 1));
                }
                if (i > 0) {
                    moves.add(swapStones(playground, i, j, i - 1, j));
                }
                if (j > 0) {
                    moves.add(swapStones(playground, i, j, i, j - 1));
                }
            }
        }
        return moves;
    }

    /**
     * Swaps Stones and returns the result of this move
     * @param playground The playground
     * @param srcX x position of the first stone
     * @param srcY y position of the first stone
     * @param dstX x position of the second stone
     * @param dstY y position of the second stone
     * @return returns a list of all results from this move
     */
    private Move swapStones(Stone[][] playground, int srcX, int srcY, int dstX, int dstY) {
        Stone[][] tmpPlayground = new Stone[10][10];
        for (int i = 0; i < 10; i++) {
            for (int j = 0; j < 10; j++) {
                tmpPlayground[i][j] = playground[i][j];
            }
        }
        Stone src = playground[srcX][srcY];
        Stone dst = playground[dstX][dstY];
        tmpPlayground[srcX][srcY] = dst;
        tmpPlayground[dstX][dstY] = src;

        ArrayList<Row> rows = new ArrayList<Row>();
        ArrayList<Row> tmpRows = analysePlayground(tmpPlayground);
        rows.addAll(tmpRows);
        int depth = 0;
        while (tmpRows.size() > 0 && depth < 10) {
            depth++;
            tmpPlayground = removeRowsFromPlayground(tmpPlayground, tmpRows);
            tmpRows = analysePlayground(tmpPlayground);
            try {
                rows.addAll(rows);
            } catch (OutOfMemoryError e) {
                break;
            }
//            Logger.getLogger(Main.class.getName()).log(Level.INFO, "Number of rows: " + tmpRows.size());
//            printPlayground(tmpPlayground);
        }

        Move m = new Move();
        m.setSrcX(srcX);
        m.setSrcY(srcY);
        m.setDstX(dstX);
        m.setDstY(dstY);
        m.setRows(rows);

        return m;
    }

    /**
     * Removes Rows from playground and let the following stones drop down.
     *
     * @param playground The playground to correct
     * @param rows The Rows to remove
     * @return the corrected playground
     */
    private Stone[][] removeRowsFromPlayground(Stone[][] playground, ArrayList<Row> rows) {
        final Stone nullStone = new Stone(Stone.nullStone);
//        Logger.getLogger(Main.class.getName()).log(Level.INFO, "Number of rows: "+rows.size());

        for (Row r : rows) {
//            Logger.getLogger(Main.class.getName()).log(Level.INFO, "Row count:" + r.getCount() + ", x:" + r.getStartX() + "-" + r.getEndX() + ", y:" + r.getStartY() + "-" + r.getEndY());

            // start is smaller then end because of construction
            for (int j = r.getStartX(); j <= r.getEndX(); j++) {
                for (int k = r.getStartY(); k <= r.getEndY(); k++) {
                    playground[j][k] = nullStone;
//                    Logger.getLogger(Main.class.getName()).log(Level.INFO, "Coordniate: "+j+","+k);
                }
            }
        }

        for (int i = 0; i < playground.length; i++) {
            for (int j = 0; j < playground.length; j++) {
                if ((playground[i][j]).isNullStone()) {
                    for (int k = i; k < playground.length; k++) {
                        if (!(playground[k][j]).isNullStone()) {
                            playground[i][j] = playground[k][j];
                            playground[k][j] = nullStone;
                            break;
                        }
                    }
                }
            }
        }

        return playground;
    }

    /**
     * Returns the rows found in the given playground
     * @param tmpPlayground the playground
     * @return An ArrayList of all found rows
     */
    private ArrayList<Row> analysePlayground(Stone[][] tmpPlayground) {

        ArrayList<Row> rows = new ArrayList<Row>();
        Stone lastStone;
        Stone s;
        int count = 1;

        // check for horizontal rows of more then 2 stones
        for (int i = 0; i < 10; i++) {
            count = 1;

            for (int j = 1; j < 10; j++) {
                lastStone = tmpPlayground[i][j - 1];
                s = tmpPlayground[i][j];

                if (s.equals(lastStone)) {
                    count++;
                } else if (count > 2) {
                    Row r = new Row();
                    r.setStartX(i);
                    r.setStartY(j - count);
                    r.setEndX(i);
                    r.setEndY(j - 1);
                    r.setColor(tmpPlayground[i][j - 1]);

                    if (Stone.isBomb(tmpPlayground[i][j - 1])) {
                        int bombScore = 0;
                        for (int k = 1; k <= count; k++) {
                            bombScore += Integer.parseInt("" + (tmpPlayground[i][j - k]).getColor());
                        }
                        count = bombScore;
                    }

                    r.setCount(count);
                    rows.add(r);

                    count = 1;
                } else {
                    count = 1;
                }
            }
        }

        // check for vertical rows of more then 2 stones
        for (int i = 0; i < 10; i++) {
            count = 1;

            for (int j = 1; j < 10; j++) {
                lastStone = tmpPlayground[j - 1][i];
                s = tmpPlayground[j][i];

                if (s.equals(lastStone)) {
                    count++;
                } else if (count > 2) {
                    Row r = new Row();
                    r.setStartY(i);
                    r.setStartX(j - count);
                    r.setEndY(i);
                    r.setEndX(j - 1);
                    r.setColor(tmpPlayground[j - 1][i]);

                    if (Stone.isBomb(tmpPlayground[j - 1][i])) {
                        int bombScore = 0;
                        for (int k = 1; k <= count; k++) {
                            bombScore += Integer.parseInt("" + (tmpPlayground[j - k][i]).getColor());
                        }
                        count = bombScore;
                    }

                    r.setCount(count);
                    rows.add(r);

                    count = 1;
                } else {
                    count = 1;
                }
            }
        }

        return rows;
    }
}
